@@ -1,59 +1,14 @@
-#' Async client for different request types
+#' @title Async client for different request types
+#' @description
+#' An async client to do many requests, each with different URLs, curl options,
+#' etc.
 #'
 #' @export
-#' @param ...,.list Any number of objects of class [HttpRequest()],
-#' must supply inputs to one of these parameters, but not both
 #' @family async
 #' @template async-deets
 #' @return An object of class `AsyncVaried` with variables and methods.
-#' Responses are returned in the order they are passed in. We print the 
-#' first 10.
-#' @details
-#' **Methods**
-#'   \describe{
-#'     \item{`request()`}{
-#'       Execute asynchronous requests
-#'       - returns: nothing, responses stored inside object,
-#'       though will print messages if you choose verbose output
-#'     }
-#'     \item{`requests()`}{
-#'       list requests
-#'       - returns: a list of `HttpRequest` objects, empty list before
-#'       requests made
-#'     }
-#'     \item{`responses()`}{
-#'       list responses
-#'       - returns: a list of `HttpResponse` objects, empty list before
-#'       requests made
-#'     }
-#'     \item{`parse(encoding = "UTF-8")`}{
-#'       parse content
-#'       - returns: character vector, empty character vector before
-#'       requests made
-#'     }
-#'     \item{`status_code()`}{
-#'       (integer) HTTP status codes
-#'       - returns: numeric vector, empty numeric vector before
-#'       requests made
-#'     }
-#'     \item{`status()`}{
-#'       (list) HTTP status objects
-#'       - returns: a list of `http_code` objects, empty list before
-#'       requests made
-#'     }
-#'     \item{`content()`}{
-#'       raw content
-#'       - returns: raw list, empty list before requests made
-#'     }
-#'     \item{`times()`}{
-#'       curl request times
-#'       - returns: list of named numeric vectors, empty list before
-#'       requests made
-#'     }
-#'   }
-#'
-#' @format NULL
-#' @usage NULL
+#' [HttpResponse] objects are returned in the order they are passed in. 
+#' We print the first 10.
 #' @examples \dontrun{
 #' # pass in requests via ...
 #' req1 <- HttpRequest$new(
@@ -90,6 +45,19 @@
 #' out$parse()
 #' ## response objects
 #' out$responses()
+#' 
+#' # use $verb() method to select http verb
+#' method <- "post"
+#' req1 <- HttpRequest$new(
+#'   url = "https://httpbin.org/post",
+#'   opts = list(verbose = TRUE),
+#'   headers = list(foo = "bar")
+#' )$verb(method)
+#' req2 <- HttpRequest$new(url = "https://httpbin.org/post")$verb(method)
+#' out <- AsyncVaried$new(req1, req2)
+#' out
+#' out$request()
+#' out$responses()
 #'
 #' # pass in requests in a list via .list param
 #' reqlist <- list(
@@ -137,12 +105,27 @@
 #' tmp$request()
 #' tmp$responses()
 #' tmp$parse("UTF-8")
+#' 
+#' # access intemediate redirect headers
+#' dois <- c("10.7202/1045307ar", "10.1242/jeb.088898", "10.1121/1.3383963")
+#' reqlist <- list(
+#'   HttpRequest$new(url = paste0("https://doi.org/", dois[1]))$get(),
+#'   HttpRequest$new(url = paste0("https://doi.org/", dois[2]))$get(),
+#'   HttpRequest$new(url = paste0("https://doi.org/", dois[3]))$get()
+#' )
+#' tmp <- AsyncVaried$new(.list = reqlist)
+#' tmp$request()
+#' tmp
+#' lapply(tmp$responses(), "[[", "response_headers_all")
 #' }
 AsyncVaried <- R6::R6Class(
-  'AsyncVaried',
-  public = list(
+  "AsyncVaried",
+  public = list(    
+    #' @description print method for AsyncVaried objects
+    #' @param x self
+    #' @param ... ignored
     print = function(x, ...) {
-      cat("<crul async varied connection> ", sep = "\n")
+      cat(private$print_string, sep = "\n")
       cat(sprintf("  requests: (n: %s)", length(private$reqs)), sep = "\n")
       print_urls <- private$reqs[1:min(c(length(private$reqs), 10))]
       for (i in seq_along(print_urls)) {
@@ -151,11 +134,16 @@ AsyncVaried <- R6::R6Class(
                     print_urls[[i]]$url), "\n")
       }
       if (length(private$reqs) > 10) {
-        cat(sprintf("   # ... with %s more", length(private$reqs) - 10), sep = "\n")
+        cat(sprintf("   # ... with %s more", length(private$reqs) - 10),
+          sep = "\n")
       }
       invisible(self)
     },
 
+    #' @description Create a new AsyncVaried object
+    #' @param ...,.list Any number of objects of class [HttpRequest()],
+    #' must supply inputs to one of these parameters, but not both
+    #' @return A new `AsyncVaried` object
     initialize = function(..., .list = list()) {
       if (length(.list)) {
         private$reqs <- .list
@@ -172,43 +160,65 @@ AsyncVaried <- R6::R6Class(
       }
     },
 
+    #' @description Execute asynchronous requests
+    #' @return nothing, responses stored inside object, though will print
+    #' messages if you choose verbose output
     request = function() {
       private$output <- private$async_request(private$reqs)
     },
 
+    #' @description List responses
+    #' @return a list of `HttpResponse` objects, empty list before
+    #' requests made
     responses = function() {
       private$output %||% list()
     },
 
+    #' @description List requests
+    #' @return a list of `HttpRequest` objects, empty list before
+    #' requests made
     requests = function() {
       private$reqs
     },
 
+    #' @description parse content
+    #' @param encoding (character) the encoding to use in parsing.
+    #' default:"UTF-8"
+    #' @return character vector, empty character vector before
+    #' requests made
     parse = function(encoding = "UTF-8") {
       vapply(private$output, function(z) z$parse(encoding = encoding), "")
     },
 
+    #' @description Get HTTP status codes for each response
+    #' @return numeric vector, empty numeric vector before requests made
     status_code = function() {
       vapply(private$output, function(z) z$status_code, 1)
     },
 
+    #' @description List HTTP status objects
+    #' @return a list of `http_code` objects, empty list before requests made
     status = function() {
       lapply(private$output, function(z) z$status_http())
     },
 
+    #' @description Get raw content for each response
+    #' @return raw list, empty list before requests made
     content = function() {
       lapply(private$output, function(z) z$content)
     },
 
+    #' @description curl request times
+    #' @return list of named numeric vectors, empty list before requests made
     times = function() {
       lapply(private$output, function(z) z$times)
     }
   ),
 
   private = list(
+    print_string = "<crul async varied connection>",
     reqs = NULL,
     output = NULL,
-    reqq = NULL,
 
     async_request = function(reqs) {
       crulpool <- curl::new_pool()
@@ -236,18 +246,26 @@ AsyncVaried <- R6::R6Class(
             ff <- file(w$disk, open = "wb")
             curl::multi_add(
               handle = h,
-              done = function(res) multi_res[[i]] <<- res,
-              fail = function(res) multi_res[[i]] <<- make_async_error(res, w),
+              done = function(res) {
+                close(ff)
+                multi_res[[i]] <<- res
+              },
+              fail = function(res) {
+                close(ff)
+                multi_res[[i]] <<- make_async_error(res, w)
+              },
               data = ff,
               pool = crulpool
             )
           } else if (is.null(w$disk) && !is.null(w$stream)) {
             stopifnot(is.function(w$stream))
+            # assign empty response since stream is a user supplied function
+            # to write somewhere of their choosing
+            multi_res[[i]] <<- make_async_error("", w)
             curl::multi_add(
               handle = h,
-              done = function(res) multi_res[[i]] <<- res,
+              done = w$stream,
               fail = function(res) multi_res[[i]] <<- make_async_error(res, w),
-              data = w$stream,
               pool = crulpool
             )
           }
@@ -263,24 +281,28 @@ AsyncVaried <- R6::R6Class(
       multi_res <- ccp(multi_res)
 
       Map(function(z, b) {
+        # prep headers
+        if (grepl("^ftp://", z$url)) {
+          headers <- list()
+        } else {
+          hh <- rawToChar(z$headers %||% raw(0))
+          if (nzchar(hh)) {
+            headers <- lapply(curl::parse_headers(hh, multiple = TRUE),
+              head_parse)
+          } else {
+            headers <- list()
+          }
+        }
         HttpResponse$new(
           method = b$payload$method,
           url = z$url,
           status_code = z$status_code,
-          request_headers = c(useragent = b$payload$options$useragent,
-                              b$headers),
-          response_headers = {
-            if (grepl("^ftp://", z$url)) {
-              list()
-            } else {
-              head_char <- rawToChar(z$headers)
-              if (nzchar(head_char)) {
-                headers_parse(curl::parse_headers(head_char))
-              } else {
-                list()
-              }
-            }
-          },
+          request_headers = c(
+            useragent = b$payload$options$useragent,
+            b$headers
+          ),
+          response_headers = last(headers),
+          response_headers_all = headers,
           modified = z$modified,
           times = z$times,
           content = z$content,
